


var lastDate
, namespaces = createDebug.namespaces = {}
, filters = []
, lastLog = {}


module.exports = createDebug

createDebug.enable  = setLogger.bind(null, true)
createDebug.disable = setLogger.bind(null, false)
createDebug.log = log


function createDebug(name, enable) {
	if (enable) {
		var re = makeRe(typeof enable == "string" ? enable : name)
		if (filters.map(String).indexOf("" + re) == -1) {
			filters.push(re)
		}
	}
	namespaces[name] = filters.some(match, name)
	return log.bind(null, name)
}

function match(re) {
	return re.test(this)
}

function log(name, msg) {
	if (!namespaces[name]) return

	var args = arguments
	, now = new Date()
	, nowParts = now.toISOString().split("T")
	, ms = now - (lastLog[name] || now)
	, sec = 1000
	, min = 60 * sec
	, hour = 60 * min
	, time = (
		ms > hour ? (ms / hour).toFixed(1) + "h" :
		ms > min ? (ms / min).toFixed(1) + "m" :
		ms > sec ? (ms / sec | 0) + "s" :
		ms + "ms"
	).replace(".0", "")

	if (lastDate !== nowParts[0]) {
		lastDate = nowParts[0]
		console.log("\n===\n" + lastDate + "\n===\n")
	}

	if (msg instanceof Error) {
		msg = msg.stack || msg.message
	}

	lastLog[name] = now

	args[1] = name
	+ " +" + time
	+ " " + msg
	args[0] = nowParts[1]


	console.log.apply(console, args)
}

function makeRe(str) {
	return new RegExp("^" + str.split("*").join(".*?") + "$")
}

function setLogger(state, filter, note) {
	if (!filter) return

	var res = filter
	.split(/[\s,]+/)
	.map(makeRe)
	.filter(function(re) {
		var idx = this.indexOf("" + re)
		if (state && idx == -1) filters.push(re)
		else if (!state && idx > -1) filters.splice(idx, 1)

		return state != (idx > -1)
	}, filters.map(String))

	Object.keys(namespaces).forEach(function(name) {
		namespaces[name] = filters.some(match, name)
	})

	log("debug", (state ? "enable[" : "disable[") + res.length + "]:" + filter + " " + note)
}



